//   Copyright (c) 2010-2012 Giovanni Lagorio and John Abbott, Anna M. Bigatti
//
//   This file is part of the source of CoCoALib, the CoCoA Library.
//
//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.

#include "BuiltInFunctions.H"

using namespace std;
using namespace boost;
using namespace boost::iostreams;
using namespace CoCoA::AST;
using namespace CoCoA::LexerNS;
using namespace CoCoA::ParserNS;

namespace CoCoA {
namespace InterpreterNS {

// builtIns is essentially a global variable but with better controlled initialization.
std::vector<NameFunPair>& builtIns()
{
  static std::vector<NameFunPair> GlobalVar;
  return GlobalVar;
}

DECLARE_STD_BUILTIN_FUNCTION(IsInteger, 2) {
	intrusive_ptr<LeftValue> resultRef = intrusive_ptr_cast<LeftValue>(runtimeEnv->evalArg(ARG(0), RuntimeEnvironment::EVAL_BY_REF));
	intrusive_ptr<RingElemValue> poly = runtimeEnv->evalArgAs<RingElemValue>(ARG(1));
	BigInt N;
	bool isInteger(IsInteger(N, poly->theRingElem));
	if (isInteger) {
		if (resultRef->assignmentNeedsOwnership())
			resultRef->obtainOwnership();
		const intrusive_ptr<const Expression> resultRefExp = ARG(0).exp;
		resultRef->assign(new BigIntValue(N), resultRefExp->getBegin(), resultRefExp->getEnd(), runtimeEnv);
	}
	return BoolValue::fromBool(isInteger);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(IsRational, 2) {
	intrusive_ptr<LeftValue> resultRef = intrusive_ptr_cast<LeftValue>(runtimeEnv->evalArg(ARG(0), RuntimeEnvironment::EVAL_BY_REF));
	intrusive_ptr<RingElemValue> poly = runtimeEnv->evalArgAs<RingElemValue>(ARG(1));
	BigRat qq;
	bool isRational(IsRational(qq, poly->theRingElem));
	if (isRational) {
		if (resultRef->assignmentNeedsOwnership())
			resultRef->obtainOwnership();
		const intrusive_ptr<const Expression> resultRefExp = ARG(0).exp;
		resultRef->assign(new BigRatValue(qq), resultRefExp->getBegin(), resultRefExp->getEnd(), runtimeEnv);
	}
	return BoolValue::fromBool(isRational);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(aliases, 0) {
	runtimeEnv->topLevelAliases->dump(runtimeEnv->getOutputStream());
	return VoidValue::theInstance;
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(pause, 0) {
#ifdef C5IDE
	runtimeEnv->interpreter->singleStepExecution = true;
#else // #ifdef C5IDE
	(void)runtimeEnv; // keeps the compiler happy (or, at least, silent ;-) )
#endif // #ifdef C5IDE
	return VoidValue::theInstance;
}
END_STD_BUILTIN_FUNCTION

#ifdef C5IDE
DECLARE_STD_BUILTIN_FUNCTION(sleep, 1) {
	const Argument &arg = ARG(0);
	intrusive_ptr<BigIntValue> N = runtimeEnv->evalArgAs<BigIntValue>(arg);
	long l;
	if (!IsConvertible(l, N->theInteger) || l<0 || l>1000)
		throw RuntimeException("The argument must be in the range [0..1000]", arg.exp);
	this_thread::sleep(boost::posix_time::seconds(static_cast<int>(l)));
	return VoidValue::theInstance;
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(debug, 1) {
	const Argument &arg = ARG(0);
	string filename = runtimeEnv->evalArgAs<StringValue>(arg)->theString;
	try {
		runtimeEnv->interpreter->singleStepExecution = true;
		runtimeEnv->interpreter->readAndExecute(filename, false, false);
	} catch (const RuntimeException &) {
		throw;
	} catch (const BaseException &be) {
		throw RuntimeException(be.reason, invocationExpression->getBegin(), invocationExpression->getEnd());
	}
	return VoidValue::theInstance;
}
END_STD_BUILTIN_FUNCTION
#endif // #ifdef C5IDE

#if 0  // 2 functions for accessing zip files, currently disabled
DECLARE_STD_BUILTIN_FUNCTION(ZipFileList, 1) {
	const Argument &arg = ARG(0);
	const string filename = runtimeEnv->evalArgAs<StringValue>(arg)->theString;
	int err;
	struct zip *zf = ::zip_open(filename.c_str(), ZIP_CHECKCONS, &err);
	if (!zf)
		throw RuntimeException("Cannot read Zip file \""+filename+"\"", arg.exp);
	intrusive_ptr<ListValue> result(new ListValue);
	const int numEntries = zip_get_num_files(zf);
	for(int a=0; a<numEntries; ++a)
		result->addValue(new StringValue(zip_get_name(zf, a, ZIP_FL_UNCHANGED)));
	::zip_close(zf);
	return result;
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(ZipRead, 2) {
	const Argument &argFilename = ARG(0);
	const Argument &argEntryName = ARG(1);
	const string filename = runtimeEnv->evalArgAs<StringValue>(argFilename)->theString;
	const string entryName = runtimeEnv->evalArgAs<StringValue>(argEntryName)->theString;
	int err;
	struct zip *zf = ::zip_open(filename.c_str(), ZIP_CHECKCONS, &err);
	if (!zf)
		throw RuntimeException("Cannot read Zip file \""+filename+"\"", argFilename.exp);
	const int index = zip_name_locate(zf, entryName.c_str(), /* flags */ 0);
	if (index<0) {
		::zip_close(zf);
		throw RuntimeException("Cannot find an entry named \""+entryName+" inside the Zip file \""+filename+"\"", argEntryName.exp);
	}
	struct zip_file *entry = ::zip_fopen_index(zf, index, /* flags */ 0);
	if (!entry) {
		::zip_close(zf);
		throw RuntimeException("Error reading the entry named \""+entryName+" inside the Zip file \""+filename+"\"", argEntryName.exp);
	}
	string result;
	for(;;) {
		char buf[1024];
		int nRead = ::zip_fread(entry, buf, sizeof(buf));
		if (nRead<=0) {
			::zip_fclose(entry);
			::zip_close(zf);
			return new StringValue(result);
		}
		result.append(buf, buf+nRead);
	}
}
END_STD_BUILTIN_FUNCTION
#endif

DECLARE_STD_BUILTIN_FUNCTION(len, 1) {
	intrusive_ptr<const RightValue> v=runtimeEnv->evalArgAs<RightValue>(ARG(0));
	if (dynamic_pointer_cast<const MatrixValue>(v))
    throw RuntimeException("len(Matrix) not allowed, use NumRows(matrix) instead", ARG(0).exp);
	if (dynamic_pointer_cast<const MatrixRowValue>(v))
    throw RuntimeException("len(MatrixRow) not allowed, use NumCols(matrix) instead", ARG(0).exp);
	int which;
	v = runtimeEnv->evalArgAsT1orT2orT3<StringValue, ListValue, RingElemValue>(ARG(0), which);
  switch (which) {
  case 1: return Value::from(BigInt(PtrCastString(v).length()));
	case 2: return Value::from(BigInt(intrusive_ptr_cast<const ListValue>(v)->size()));
  default: return Value::from(BigInt(NumTerms(PtrCastRingElem(v))));
  }
}
END_STD_BUILTIN_FUNCTION


DECLARE_ARITYCHECK_FUNCTION(sum) { return (1<=nArg) && (nArg<=2); }
DECLARE_BUILTIN_FUNCTION(sum) {
	invocationExpression->checkNumberOfArgs(1,2);
	const Argument &a0 = ARG(0);
	intrusive_ptr<ListValue> list = runtimeEnv->evalArgAs<ListValue>(a0);
	ListValue::ContainerType::size_type size = list->size();
  intrusive_ptr<RightValue> result = BigIntValue::zero;
  if (invocationExpression->args.size()==1)
  {
    if (size==0) return result;
    result = list->getValue(--size);
  }
  else
    result = runtimeEnv->evalArgAs<RightValue>(ARG(1));
  for(ListValue::ContainerType::size_type a=size; a!=0; /**/)
    try	{
			result = runtimeEnv->binaryOperatorDispatch(list->getValue(--a), result, RuntimeEnvironment::opPlusMap, a0.exp->getBegin(), a0.exp->getEnd());
	} catch (const InterruptException &) {
		throw;
	} catch (const RuntimeException &) {
		throw RuntimeException("Some elements have incompatible types for the sum", a0.exp);
	}
	return result;
}

DECLARE_ARITYCHECK_FUNCTION(product) { return (1<=nArg) && (nArg<=2); }
DECLARE_BUILTIN_FUNCTION(product) {
	invocationExpression->checkNumberOfArgs(1,2);
	const Argument &a0 = ARG(0);
	intrusive_ptr<ListValue> list = runtimeEnv->evalArgAs<ListValue>(a0);
	ListValue::ContainerType::size_type size = list->size();
	intrusive_ptr<RightValue> result = BigIntValue::one;
  if (invocationExpression->args.size()==1)
  {
    if (size==0) return result;
    result = list->getValue(--size);
  }
  else
    result = runtimeEnv->evalArgAs<RightValue>(ARG(1));
  for(ListValue::ContainerType::size_type a=size; a!=0; /**/)
		try	{
			result = runtimeEnv->binaryOperatorDispatch(list->getValue(--a), result, RuntimeEnvironment::opStarMap, a0.exp->getBegin(), a0.exp->getEnd());
	} catch (const InterruptException &) {
		throw;
	} catch (const RuntimeException &) {
		throw RuntimeException("Some elements have incompatible types for the product", a0.exp);
	}
	return result;
}

DECLARE_ARITYCHECK_FUNCTION(NewList) { return (1<=nArg) && (nArg<=2); }
DECLARE_BUILTIN_FUNCTION(NewList) {
	invocationExpression->checkNumberOfArgs(1,2);
	intrusive_ptr<BigIntValue> N = runtimeEnv->evalArgAs<BigIntValue>(ARG(0));
  long n;
  if (!IsConvertible(n, N->theInteger) || n<0)
     throw RuntimeException("invalid length", ARG(0).exp);
	intrusive_ptr<ListValue> returnValue = new ListValue();
  intrusive_ptr<RightValue> val = BigIntValue::zero;
  if (invocationExpression->args.size()==2)
    val = runtimeEnv->evalArgAs<RightValue>(ARG(1));
  for(long i=0; i<n; ++i)  returnValue->addValue(val);
	return returnValue;
}

DECLARE_ARITYCHECK_FUNCTION(first) { return (1<=nArg) && (nArg<=2); }
DECLARE_BUILTIN_FUNCTION(first) {
	invocationExpression->checkNumberOfArgs(1,2);
	intrusive_ptr<ListValue> list = runtimeEnv->evalArgAs<ListValue>(ARG(0));
  if (invocationExpression->args.size()==1)
  {
    if (list->size() == 0) throw RuntimeException("list is empty", ARG(0).exp);
    return list->getValue(0);
  }
	intrusive_ptr<BigIntValue> N = runtimeEnv->evalArgAs<BigIntValue>(ARG(1));
  long n;
  if (!IsConvertible(n, N->theInteger) || n<0)
    throw RuntimeException("invalid length", ARG(1).exp);
  if ((unsigned long)n > list->size())
    throw RuntimeException("value greater than list length", ARG(1).exp);
	intrusive_ptr<ListValue> returnValue = new ListValue();
	for(long i=0; i<n; ++i)  returnValue->addValue(list->getValue(i));
	return returnValue;
}

DECLARE_ARITYCHECK_FUNCTION(last) { return (1<=nArg) && (nArg<=2); }
DECLARE_BUILTIN_FUNCTION(last) {
	invocationExpression->checkNumberOfArgs(1,2);
	intrusive_ptr<ListValue> list = runtimeEnv->evalArgAs<ListValue>(ARG(0));
  ///////// here
  ListValue::ContainerType::size_type s = list->size();
  if (invocationExpression->args.size()==1) 
  if (invocationExpression->args.size()==1)
  {
    if (list->size() == 0) throw RuntimeException("list is empty", ARG(0).exp);
    return list->getValue(s-1);
  }
	intrusive_ptr<BigIntValue> N = runtimeEnv->evalArgAs<BigIntValue>(ARG(1));
  long n;
  if (!IsConvertible(n, N->theInteger) || n<0)
    throw RuntimeException("invalid length", ARG(1).exp);
  if ((unsigned long)n>s)
    throw RuntimeException("value greater than list length", ARG(1).exp);
	intrusive_ptr<ListValue> returnValue = new ListValue();
	for(unsigned long i=s-n; i<s; ++i)  returnValue->addValue(list->getValue(i));
	return returnValue;
}

DECLARE_STD_BUILTIN_FUNCTION(ScalarProduct, 2) {
	const Argument &arg1 = ARG(0);
	intrusive_ptr<ListValue> l1 = runtimeEnv->evalArgAs<ListValue>(arg1);
	const Argument &arg2 = ARG(1);
	intrusive_ptr<ListValue> l2 = runtimeEnv->evalArgAs<ListValue>(arg2);
	const ListValue::ContainerType::size_type size = l1->size();
	const CharPointer &begin = arg1.exp->getBegin();
	const CharPointer &end = arg2.exp->getEnd();
	if (l2->size()!=size)
		throw RuntimeException("The arguments must be lists of the same size", begin, end);
	intrusive_ptr<RightValue> result = BigIntValue::zero;
	try	{
		for(ListValue::ContainerType::size_type a=0; a<size; ++a)
			result = runtimeEnv->binaryOperatorDispatch(
						result,
						runtimeEnv->binaryOperatorDispatch(l1->getValue(a), l2->getValue(a), RuntimeEnvironment::opStarMap, begin, end),
						RuntimeEnvironment::opPlusMap,
						begin,
						end);
	} catch (const InterruptException &) {
		throw;
	} catch (const RuntimeException &) {
		throw RuntimeException("Some elements have incompatible types for the scalar product", begin, end);
	}
	return result;
}
END_STD_BUILTIN_FUNCTION

DECLARE_ARITYCHECK_FUNCTION(concat) { (void)nArg; return true; }

DECLARE_BUILTIN_FUNCTION(concat) {
	intrusive_ptr<ListValue> returnValue = new ListValue();
	BOOST_FOREACH(const Argument &arg, invocationExpression->args) {
		intrusive_ptr<ListValue> list = runtimeEnv->evalArgAs<ListValue>(arg);
		int size = list->size();
		for(int a=0; a<size; ++a)
			returnValue->addValue(list->getValue(a));
	}
	return returnValue;
}

DECLARE_STD_BUILTIN_FUNCTION(reversed, 1) {
	intrusive_ptr<ListValue> list = runtimeEnv->evalArgAs<ListValue>(ARG(0));
	intrusive_ptr<ListValue> returnValue = new ListValue();
	for(ListValue::ContainerType::size_type a=list->size(); a!=0; /**/)
		returnValue->addValue(list->getValue(--a));
	return returnValue;
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(reverse, 1) {
	intrusive_ptr<ListValue> list = runtimeEnv->obtainUnshared<ListValue>(ARG(0));
	const ListValue::ContainerType::size_type size = list->size();
	if (size) {
		ListValue::ContainerType::size_type left = 0, right = size-1;
		while(left < right) {
			const intrusive_ptr<RightValue> tmp(list->getValue(left));
			list->setValue(left++, list->getValue(right));
			list->setValue(right--, tmp);
		}
	}
	return VoidValue::theInstance;
}
END_STD_BUILTIN_FUNCTION

DECLARE_ARITYCHECK_FUNCTION(deg) { return (1<=nArg) && (nArg<=2); }
DECLARE_BUILTIN_FUNCTION(deg) {  // AMB
	invocationExpression->checkNumberOfArgs(1,2);
	intrusive_ptr<RingElemValue> a = runtimeEnv->evalArgAs<RingElemValue>(ARG(0));
  if (invocationExpression->args.size()==1)
    return Value::from(deg(a->theRingElem));
	intrusive_ptr<RingElemValue> b = runtimeEnv->evalArgAs<RingElemValue>(ARG(1));
  long i;
  if (!IsIndet(i,b->theRingElem))
    throw RuntimeException("not an indet", ARG(1).exp);
  return Value::from(MaxExponent(a->theRingElem, i));
}

// variable number of args
DECLARE_ARITYCHECK_FUNCTION(coefficients) { return (1<=nArg) && (nArg<=2); }
DECLARE_BUILTIN_FUNCTION(coefficients) {  // AMB
	invocationExpression->checkNumberOfArgs(1,2);
	intrusive_ptr<RingElemValue> a = runtimeEnv->evalArgAs<RingElemValue>(ARG(0));
  if (invocationExpression->args.size()==1)
    return Value::from(coefficients_forC5(a->theRingElem));
	vector<RingElem> v = runtimeEnv->evalArgAsRingElemList(ARG(1));
  vector<RingElem> res;
  for (long i=0; i<len(v); ++i)
    res.push_back(CoeffOfTerm_forC5(a->theRingElem, v[i]));
  return Value::from(res);
}

// variable number of args
DECLARE_ARITYCHECK_FUNCTION(NewMat) { return (2<=nArg) && (nArg<=3); }
DECLARE_BUILTIN_FUNCTION(NewMat) {
	invocationExpression->checkNumberOfArgs(2,3);
  if (invocationExpression->args.size()==2)
    throw RuntimeException("NewMat(R,C) not allowed, use NewMat(Ring,R,C) or NewMat(R,C,Entry) instead", ARG(0).exp);
  long nrows, ncols;
	intrusive_ptr<BigIntValue> nrows_ptr, ncols_ptr;
	intrusive_ptr<RightValue> v0 = runtimeEnv->evalArgAs<RightValue>(ARG(0));
	intrusive_ptr<RightValue> v1 = runtimeEnv->evalArgAs<RightValue>(ARG(1));
	intrusive_ptr<RightValue> v2 = runtimeEnv->evalArgAs<RightValue>(ARG(2));
	if (const intrusive_ptr<const RingValue> R = dynamic_pointer_cast<const RingValue>(v0))
  {
    if ( !(nrows_ptr = dynamic_pointer_cast<BigIntValue>(v1)) 
         || !IsConvertible(nrows, nrows_ptr->theInteger)
         || nrows<0)
      throw RuntimeException("invalid number of rows", ARG(1).exp);
    if ( !(ncols_ptr = dynamic_pointer_cast<BigIntValue>(v2)) 
         || !IsConvertible(ncols, ncols_ptr->theInteger)
         || ncols<0)
      throw RuntimeException("invalid number of columns", ARG(2).exp);
    return Value::from(ZeroMat(R->theRing, nrows, ncols));
  }
    if ( !(nrows_ptr = dynamic_pointer_cast<BigIntValue>(v0)) 
         || !IsConvertible(nrows, nrows_ptr->theInteger)
         || nrows<0)
      throw RuntimeException("invalid number of rows", ARG(0).exp);
    if ( !(ncols_ptr = dynamic_pointer_cast<BigIntValue>(v1)) 
         || !IsConvertible(ncols, ncols_ptr->theInteger)
         || ncols<0)
      throw RuntimeException("invalid number of columns", ARG(1).exp);
	if (intrusive_ptr<RingElemValue> x = dynamic_pointer_cast<RingElemValue>(v2))
    return Value::from(FilledMat(owner(x->theRingElem), nrows, ncols, x->theRingElem));
	if (intrusive_ptr<BigIntValue> x = dynamic_pointer_cast<BigIntValue>(v2))
    return Value::from(FilledMat(RingQQ(), nrows, ncols, RingElem(RingQQ(), x->theInteger)));
	if (intrusive_ptr<BigRatValue> x = dynamic_pointer_cast<BigRatValue>(v2))
    return Value::from(FilledMat(RingQQ(), nrows, ncols, RingElem(RingQQ(), x->theRational)));
	throw WrongTypeException(RingElemValue::type->name + " or " + 
                           BigIntValue::type->name + " or " + 
                           BigRatValue::type->name, v2->getType()->name, ARG(2).exp);
}

// variable number of args
DECLARE_ARITYCHECK_FUNCTION(jacobian) { return (1<=nArg) && (nArg<=2); }
DECLARE_BUILTIN_FUNCTION(jacobian) {  // AMB
	invocationExpression->checkNumberOfArgs(1,2);
	vector<RingElem> v0 = runtimeEnv->evalArgAsRingElemList(ARG(0));
  if (invocationExpression->args.size()==1)
  {
    if (v0.size()==0) throw RuntimeException("Empty list", ARG(0).exp);
    return new MatrixValue(jacobian(v0, indets(AsPolyRing(owner(v0[0])))));
  }
	vector<RingElem> v1 = runtimeEnv->evalArgAsRingElemList(ARG(1));
  return new MatrixValue(jacobian(v0, v1));
}

// variable number of args
DECLARE_ARITYCHECK_FUNCTION(CompleteToOrd) { return (1<=nArg) && (nArg<=2); }
DECLARE_BUILTIN_FUNCTION(CompleteToOrd) {  // AMB
	invocationExpression->checkNumberOfArgs(1,2);
	intrusive_ptr<const MatrixValue> M0 = runtimeEnv->evalArgAs<const MatrixValue>(ARG(0));
  if (invocationExpression->args.size()==1)
    return Value::from(NewMatCompleteOrd(M0->theMatrix));
	intrusive_ptr<const MatrixValue> M1 = runtimeEnv->evalArgAs<const MatrixValue>(ARG(1));
  return Value::from(NewMatCompleteOrd(M0->theMatrix, M1->theMatrix));
}


DECLARE_STD_BUILTIN_FUNCTION(HomogElimMat, 2) { // AMB
	intrusive_ptr<MatrixValue> M = runtimeEnv->evalArgAs<MatrixValue>(ARG(0));
	vector<BigInt> v1 = runtimeEnv->evalArgAsListOf<BigIntValue>(ARG(1));
  return Value::from(HomogElimMat_forC5(M->theMatrix, v1));
}
END_STD_BUILTIN_FUNCTION


// DECLARE_ARITYCHECK_FUNCTION(GrammSchmidtRows) { return (1<=nArg) && (nArg<=2); }
// DECLARE_BUILTIN_FUNCTION(GrammSchmidtRows) {  // AMB
// 	invocationExpression->checkNumberOfArgs(1,2);
// 	intrusive_ptr<MatrixValue> M = runtimeEnv->obtainUnshared<MatrixValue>(ARG(0));
//   if (invocationExpression->args.size()==1)
//   {
//     GrammSchmidtRows(M->theMatrix);
//     return VoidValue::theInstance;
//   }
// 	intrusive_ptr<BigIntValue> N = runtimeEnv->evalArgAs<BigIntValue>(ARG(1));
//   long n;
//   if (!IsConvertible(n, N->theInteger))
//     throw RuntimeException("invalid row index", ARG(1).exp);
//   GrammSchmidtRows(M->theMatrix, n);
//   return VoidValue::theInstance;
// }


DECLARE_STD_BUILTIN_FUNCTION(append, 2) {
	intrusive_ptr<RightValue> elem = runtimeEnv->evalArgAs<RightValue>(ARG(1));
	intrusive_ptr<ListValue> list = runtimeEnv->obtainUnshared<ListValue>(ARG(0));
	list->addValue(elem);
	return VoidValue::theInstance;
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(remove, 2) {
	const Argument indexArg = ARG(1);
	intrusive_ptr<RightValue> shouldBeAnIndex = runtimeEnv->evalArgAs<RightValue>(indexArg);
	intrusive_ptr<ListValue> list = runtimeEnv->obtainUnshared<ListValue>(ARG(0));
	ListValue::ContainerType::size_type index = list->checkIndex(shouldBeAnIndex, indexArg.exp);
	list->removeValue(index);
	return VoidValue::theInstance;
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(insert, 3) {
	const Argument &indexArg = ARG(1);
	intrusive_ptr<RightValue> shouldBeAnIndex = runtimeEnv->evalArgAs<RightValue>(indexArg);
	intrusive_ptr<RightValue> elem = runtimeEnv->evalArgAs<RightValue>(ARG(2));
	intrusive_ptr<ListValue> list = runtimeEnv->obtainUnshared<ListValue>(ARG(0));
	ListValue::ContainerType::size_type index = list->checkIndex(shouldBeAnIndex, indexArg.exp);
	list->insertValue(index, elem);
	return VoidValue::theInstance;
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(div, 2) {
	intrusive_ptr<BigIntValue> a = runtimeEnv->evalArgAs<BigIntValue>(ARG(0));
	intrusive_ptr<BigIntValue> b = runtimeEnv->evalArgAs<BigIntValue>(ARG(1));
	BigInt result;
	div(result, a->theInteger, b->theInteger);
	return new BigIntValue(result);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(mod, 2) {
	intrusive_ptr<BigIntValue> a = runtimeEnv->evalArgAs<BigIntValue>(ARG(0));
	intrusive_ptr<BigIntValue> b = runtimeEnv->evalArgAs<BigIntValue>(ARG(1));
	BigInt result;
	mod(result, a->theInteger, b->theInteger);
	return new BigIntValue(result);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(NumDigits, 2) {
	intrusive_ptr<BigIntValue> a = runtimeEnv->evalArgAs<BigIntValue>(ARG(0));
	intrusive_ptr<BigIntValue> b = runtimeEnv->evalArgAs<BigIntValue>(ARG(1));
	long base;
	if (!IsConvertible(base, b->theInteger) || base<2 || base>36)
		throw RuntimeException("Base must be in the range 2..36", ARG(1).exp);
	return new BigIntValue(NumDigits(a->theInteger, base));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(CRT, 4) {
	intrusive_ptr<BigIntValue> R1 = runtimeEnv->evalArgAs<BigIntValue>(ARG(0));
	intrusive_ptr<BigIntValue> M1 = runtimeEnv->evalArgAs<BigIntValue>(ARG(1));
	intrusive_ptr<BigIntValue> R2 = runtimeEnv->evalArgAs<BigIntValue>(ARG(2));
	intrusive_ptr<BigIntValue> M2 = runtimeEnv->evalArgAs<BigIntValue>(ARG(3));
        CRTMill CRT;
        CRT.myAddInfo(R1->theInteger, M1->theInteger);
        CRT.myAddInfo(R2->theInteger, M2->theInteger);

        // Create return value: ... record
        intrusive_ptr<RecordValue> ans(new RecordValue);
        ans->setField("Residue", Value::from(residue(CRT)));
        ans->setField("Modulus", Value::from(modulus(CRT)));
        return ans;
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(RatReconstructWithBounds, 5) {
	intrusive_ptr<BigIntValue> E = runtimeEnv->evalArgAs<BigIntValue>(ARG(0));
	intrusive_ptr<BigIntValue> P = runtimeEnv->evalArgAs<BigIntValue>(ARG(1));
	intrusive_ptr<BigIntValue> Q = runtimeEnv->evalArgAs<BigIntValue>(ARG(2));
	const vector<BigInt> ResList = runtimeEnv->evalArgAsListOf<BigIntValue>(ARG(3));
	const vector<BigInt> ModList = runtimeEnv->evalArgAsListOf<BigIntValue>(ARG(4));

        const long NumRes = len(ResList);
        vector<long> res(NumRes);
        for (long i=0; i < NumRes; ++i)
          res[i] = ConvertTo<long>(ResList[i]);
        const long NumMod = len(ModList);
        vector<long> mod(NumMod);
        for (long i=0; i < NumMod; ++i)
          mod[i] = ConvertTo<long>(ModList[i]);
        const long e = ConvertTo<long>(E->theInteger);
        BigRat ans = RatReconstructWithBounds(e, P->theInteger, Q->theInteger, res, mod);
        if (den(ans) == 0) return Value::from(string("FAILURE"));
        return Value::from(ans);
}
END_STD_BUILTIN_FUNCTION

DECLARE_ARITYCHECK_FUNCTION(RatReconstructByContFrac) { return (2<=nArg) && (nArg<=3); }
DECLARE_BUILTIN_FUNCTION(RatReconstructByContFrac) {
  invocationExpression->checkNumberOfArgs(2,3);
	intrusive_ptr<BigIntValue> X = runtimeEnv->evalArgAs<BigIntValue>(ARG(0));
	intrusive_ptr<BigIntValue> M = runtimeEnv->evalArgAs<BigIntValue>(ARG(1));
        // Determine threshold: 0 means use default value
        BigInt threshold; // initially 0
        if (invocationExpression->args.size()==3)
        {
          intrusive_ptr<BigIntValue> thresh = runtimeEnv->evalArgAs<BigIntValue>(ARG(2));
//          if (thresh->theInteger < 0) throw RuntimeException("Threshold must be >= 0", ARG(2).exp);
          threshold = thresh->theInteger;
        }

        RatReconstructByContFrac reconstructor(threshold);
        reconstructor.myAddInfo(X->theInteger, M->theInteger);

        // Create return value: ... record
        intrusive_ptr<RecordValue> ans(new RecordValue);
        ans->setField("IsConvincing", Value::from(IsConvincing(reconstructor)));
        if (IsConvincing(reconstructor))
          ans->setField("ReconstructedRat", Value::from(ReconstructedRat(reconstructor)));
        return ans;
}

DECLARE_ARITYCHECK_FUNCTION(RatReconstructByLattice) { return (2<=nArg) && (nArg<=3); }
DECLARE_BUILTIN_FUNCTION(RatReconstructByLattice) {
  invocationExpression->checkNumberOfArgs(2,3);
	intrusive_ptr<BigIntValue> X = runtimeEnv->evalArgAs<BigIntValue>(ARG(0));
	intrusive_ptr<BigIntValue> M = runtimeEnv->evalArgAs<BigIntValue>(ARG(1));
        // Determine threshold: 0 means use default value
        BigInt SafetyFactor; // initially 0
        if (invocationExpression->args.size()==3)
        {
          intrusive_ptr<BigIntValue> safety = runtimeEnv->evalArgAs<BigIntValue>(ARG(2));
//          if (safety->theInteger < 0) throw RuntimeException("SafetyFactor must be >= 0", ARG(2).exp);
          SafetyFactor = safety->theInteger;
        }

        RatReconstructByLattice reconstructor(SafetyFactor);
        reconstructor.myAddInfo(X->theInteger, M->theInteger);

        // Create return value: ... record
        intrusive_ptr<RecordValue> ans(new RecordValue);
        ans->setField("IsConvincing", Value::from(IsConvincing(reconstructor)));
        if (IsConvincing(reconstructor))
          ans->setField("ReconstructedRat", Value::from(ReconstructedRat(reconstructor)));
        return ans;
}

DECLARE_STD_BUILTIN_FUNCTION(sign, 1) {
	int which;
	intrusive_ptr<RightValue> v = runtimeEnv->evalArgAsT1orT2orT3<BigIntValue, BigRatValue, RingElemValue>(ARG(0), which);
  switch (which) {
  case 1: return BigIntValue::fromInt(sign(PtrCastBigInt(v)));
  case 2: return BigIntValue::fromInt(sign(PtrCastBigRat(v)));
  default:return BigIntValue::fromInt(sign(PtrCastRingElem(v)));
  }
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(IsZero, 1) {
	int which;
	intrusive_ptr<RightValue> v = runtimeEnv->evalArgAsT1orT2orT3orT4orT5<BigIntValue, BigRatValue, RingElemValue, ModuleElemValue, IdealValue>(ARG(0), which);
  switch (which) {
  case 1: return Value::from(IsZero(PtrCastBigInt(v)));
  case 2: return Value::from(IsZero(PtrCastBigRat(v)));
  case 3: return Value::from(IsZero(PtrCastRingElem(v)));
  case 4: return Value::from(IsZero(PtrCastModuleElem(v)));
  default:return Value::from(IsZero(PtrCastIdeal(v)));
  }
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(IsOne, 1) {
	int which;
	intrusive_ptr<RightValue> v = runtimeEnv->evalArgAsT1orT2orT3<BigIntValue, BigRatValue, RingElemValue>(ARG(0), which);
  switch (which) {
  case 1: return Value::from(IsOne(PtrCastBigInt(v)));
  case 2: return Value::from(IsOne(PtrCastBigRat(v)));
  default:return Value::from(IsOne(PtrCastRingElem(v)));
  }
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(IsMinusOne, 1) {
	int which;
	intrusive_ptr<RightValue> v = runtimeEnv->evalArgAsT1orT2orT3<BigIntValue, BigRatValue, RingElemValue>(ARG(0), which);
  switch (which) {
  case 1: return Value::from(IsMinusOne(PtrCastBigInt(v)));
  case 2: return Value::from(IsMinusOne(PtrCastBigRat(v)));
  default:return Value::from(IsMinusOne(PtrCastRingElem(v)));
  }
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(NumTerms, 1) { // AMB
	return Value::from(BigInt(NumTerms(runtimeEnv->evalArgAs<RingElemValue>(ARG(0))->theRingElem)));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(zero, 1) {
	bool isFirst;
  intrusive_ptr<RightValue> x = runtimeEnv->evalArgAsT1orT2<RingValue, ModuleValue>(ARG(0), isFirst);
	if (isFirst)
    return Value::from(zero(PtrCastRing(x)));
  return Value::from(zero(PtrCastModule(x)));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(gens, 1) {
	bool isFirst;
  intrusive_ptr<RightValue> x = runtimeEnv->evalArgAsT1orT2<IdealValue, ModuleValue>(ARG(0), isFirst);
	if (isFirst)
    return Value::from(gens(PtrCastIdeal(x)));
  return Value::from(gens(AsFGModule(PtrCastModule(x))));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(IsContained, 2) {
	bool isFirst;
  intrusive_ptr<RightValue> x = runtimeEnv->evalArgAsT1orT2<IdealValue, ModuleValue>(ARG(0), isFirst);
	if (isFirst)
  {
    intrusive_ptr<IdealValue> I = runtimeEnv->evalArgAs<IdealValue>(ARG(1));
    return Value::from(IsContained(PtrCastIdeal(x), I->theIdeal));
  }
  intrusive_ptr<ModuleValue> M = runtimeEnv->evalArgAs<ModuleValue>(ARG(1));
  return Value::from(IsContained(PtrCastModule(x), M->theModule));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(IsElem, 2) {
	bool isFirst;
  intrusive_ptr<RightValue> x = runtimeEnv->evalArgAsT1orT2<RingElemValue, ModuleElemValue>(ARG(0), isFirst);
	if (isFirst)
  {
    intrusive_ptr<IdealValue> I = runtimeEnv->evalArgAs<IdealValue>(ARG(1));
    return Value::from(IsElem(PtrCastRingElem(x), I->theIdeal));
  }
  intrusive_ptr<ModuleValue> M = runtimeEnv->evalArgAs<ModuleValue>(ARG(1));
  return Value::from(IsElem(PtrCastModuleElem(x), M->theModule));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(GBasis, 1) {
	bool isFirst;
  intrusive_ptr<RightValue> x = runtimeEnv->evalArgAsT1orT2<IdealValue, ModuleValue>(ARG(0), isFirst);
	if (isFirst)
    return Value::from(GBasis(PtrCastIdeal(x)));
  return Value::from(TidyGens(AsFGModule(PtrCastModule(x))));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(syz, 1) {
	int which;
	intrusive_ptr<RightValue> v = runtimeEnv->evalArgAsT1orT2orT3<IdealValue, ModuleValue, ListValue>(ARG(0), which);
  switch (which) {
  case 1: CoCoA_ERROR(ERR::NYI, "minimal syzigies");
    return Value::from(NewFreeModule(RingZZ(),1));
  case 2: CoCoA_ERROR(ERR::NYI, "minimal syzigies");
    return Value::from(NewFreeModule(RingZZ(),1));
  default:
    vector<RingElem> x = runtimeEnv->evalArgAsRingElemList(ARG(0));
    return Value::from(SyzOfGens(ideal(x)));
  }
}
END_STD_BUILTIN_FUNCTION

// variable number of args
DECLARE_ARITYCHECK_FUNCTION(SyzOfGens) { return (1<=nArg) && (nArg<=2); }
DECLARE_BUILTIN_FUNCTION(SyzOfGens) { // AMB variable number of args
	invocationExpression->checkNumberOfArgs(1,2); //variable number of args
  long n=0;
  if (invocationExpression->args.size()==2) n=1;
	bool isFirst;
  intrusive_ptr<RightValue> x = runtimeEnv->evalArgAsT1orT2<IdealValue, ModuleValue>(ARG(n), isFirst);
  if (n==0)
  {
    if (isFirst)
      return Value::from(SyzOfGens(PtrCastIdeal(x)));
    return Value::from(SyzOfGens(AsFGModule(PtrCastModule(x))));
    //    CoCoA_ERROR(ERR::NYI, "syzigies for module");
  }
  intrusive_ptr<ModuleValue> M = runtimeEnv->evalArgAs<ModuleValue>(ARG(0));
  if (isFirst)
    return Value::from(SyzOfGens(AsFreeModule(M->theModule),PtrCastIdeal(x)));
  return Value::from(SyzOfGens(AsFreeModule(M->theModule),AsFGModule(PtrCastModule(x))));
  //  CoCoA_ERROR(ERR::NYI, "syzigies for module");
  return Value::from(NewFreeModule(RingZZ(),1));
}
// END_STD_BUILTIN_FUNCTION no: variable number of args

DECLARE_STD_BUILTIN_FUNCTION(RingElem, 2) {
	intrusive_ptr<const RingValue> R(runtimeEnv->evalArgAs<RingValue>(ARG(0)));
	int which;
	intrusive_ptr<RightValue> v = runtimeEnv->evalArgAsT1orT2orT3orT4orT5<BigIntValue, BigRatValue, StringValue, RingElemValue, ListValue>(ARG(1), which);
  switch (which) {
  case 1: return Value::from(RingElem(R->theRing, PtrCastBigInt(v)));
  case 2: return Value::from(RingElem(R->theRing, PtrCastBigRat(v)));
  case 3: return Value::from(RingElem(R->theRing, symbol(PtrCastString(v))));
  case 4: return Value::from(RingElem(R->theRing, PtrCastRingElem(v)));
  default:
    const intrusive_ptr<const ListValue> l = dynamic_pointer_cast<const ListValue>(v);
    ListValue::ContainerType::size_type size = l->size();
    string str;
		if (const boost::intrusive_ptr<StringValue> s = boost::dynamic_pointer_cast<StringValue>(l->getValue(0)))
			str = s->theString;
		else
			throw RuntimeException("String expected", ARG(1).exp);
    std::vector<BigInt> indices;
    for(ListValue::ContainerType::size_type a=1; a<size; ++a) {
      if (const boost::intrusive_ptr<BigIntValue> elem = boost::dynamic_pointer_cast<BigIntValue>(l->getValue(a)))
        indices.push_back(theValue(elem));
      else
        throw RuntimeException("Integer expected", ARG(1).exp);
    }
    return Value::from(RingElem(R->theRing, symbol(str, VectorLong(indices, "RingElem"))));
  }
}
END_STD_BUILTIN_FUNCTION


DECLARE_STD_BUILTIN_FUNCTION(homog, 2) {
	int which;
	intrusive_ptr<RightValue> v = runtimeEnv->evalArgAsT1orT2orT3orT4orT5<RingElemValue, ModuleElemValue, IdealValue, ModuleValue, ListValue>(ARG(0), which);
	intrusive_ptr<const RingElemValue> h(runtimeEnv->evalArgAs<RingElemValue>(ARG(1)));
  switch (which) {
  case 1: return Value::from(homog(PtrCastRingElem(v), h->theRingElem));
  case 2: return Value::from(homog(PtrCastModuleElem(v), h->theRingElem));
  case 3: return Value::from(homog(PtrCastIdeal(v), h->theRingElem));
  case 4: CoCoA_ERROR(ERR::NYI, "homog(module)");
//return Value::from(homog(intrusive_ptr_cast<ModuleValue>(v)->theModule, h->theRingElem));
  default:
    vector<RingElem> v1 = runtimeEnv->evalArgAsRingElemList(ARG(0));
    intrusive_ptr<ListValue> returnValue = new ListValue();
    for(long a=0; a<len(v1); ++a)
      returnValue->addValue(Value::from(homog(v1[a], h->theRingElem)));
    return returnValue;
  }
}
END_STD_BUILTIN_FUNCTION


DECLARE_STD_BUILTIN_FUNCTION(abs, 1) {
	int which;
	intrusive_ptr<RightValue> v = runtimeEnv->evalArgAsT1orT2orT3<BigIntValue, BigRatValue, RingElemValue>(ARG(0), which);
  switch (which) {
  case 1: return Value::from(abs(PtrCastBigInt(v)));
  case 2: return Value::from(abs(PtrCastBigRat(v)));
  default:return Value::from(abs(PtrCastRingElem(v)));
  }
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(floor, 1) {
	bool isFirst;
  intrusive_ptr<RightValue> v = runtimeEnv->evalArgAsT1orT2<BigIntValue, BigRatValue>(ARG(0), isFirst);
	if (isFirst) return v;
	return Value::from(floor(PtrCastBigRat(v)));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(ceil, 1) {
	bool isFirst;
	intrusive_ptr<RightValue> v = runtimeEnv->evalArgAsT1orT2<BigIntValue, BigRatValue>(ARG(0), isFirst);
	if (isFirst) return v;
	return Value::from(ceil(PtrCastBigRat(v)));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(round, 1) {
	bool isFirst;
	intrusive_ptr<RightValue> v = runtimeEnv->evalArgAsT1orT2<BigIntValue, BigRatValue>(ARG(0), isFirst);
	if (isFirst) return v;
	return Value::from(round(PtrCastBigRat(v)));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(fields, 1) {
	intrusive_ptr<RecordValue> r = runtimeEnv->evalArgAs<RecordValue>(ARG(0));
	return r->fieldNames();
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(ConcatList, 1) {
	intrusive_ptr<ListValue> listOfList = runtimeEnv->evalArgAs<ListValue>(ARG(0));
	intrusive_ptr<ListValue> result(new ListValue);
	const ListValue::ContainerType::size_type nLists = listOfList->size();
	for(ListValue::ContainerType::size_type a=0; a<nLists; ++a) {
		intrusive_ptr<ListValue> l = dynamic_pointer_cast<ListValue>(listOfList->getValue(a));
		if (!l)
			throw RuntimeException("The argument is not a list of lists", ARG(0).exp);
		const ListValue::ContainerType::size_type len = l->size();
		for(ListValue::ContainerType::size_type b=0; b<len; ++b)
			result->addValue(l->getValue(b));
	}
	return result;
}
END_STD_BUILTIN_FUNCTION


namespace{ // anonymous  =============== IsVectorBigInt =================
  bool IsVectorBigInt(std::vector<BigInt>& BigIntVec, const intrusive_ptr<ListValue> l)
  {
    vector<BigInt> v;
    ListValue::ContainerType::size_type size = l->size();
    for (unsigned long i=0; i<size; ++i)
      if (const boost::intrusive_ptr<BigIntValue> n = boost::dynamic_pointer_cast<BigIntValue>(l->getValue(i)))
        v.push_back(n->theInteger);
      else
        return false;
    swap(v, BigIntVec);
    return true;
  }
}

// variable number of args
DECLARE_ARITYCHECK_FUNCTION(gcd) { return (1<=nArg) && (nArg<=2); }
DECLARE_BUILTIN_FUNCTION(gcd) {
  invocationExpression->checkNumberOfArgs(1,2);
  if (invocationExpression->args.size()==1)
  {
    intrusive_ptr<RightValue> v = runtimeEnv->evalArgAs<RightValue>(ARG(0));
    if (intrusive_ptr<ListValue> l = dynamic_pointer_cast<ListValue>(v)) {
      if (l->size() == 0) return BigIntValue::zero;
      vector<BigInt> v1;
      if (IsVectorBigInt(v1, l)) return Value::from(gcd_forC5(v1));
      return Value::from(gcd_forC5(runtimeEnv->evalArgAsRingElemList(ARG(0))));
    }
    if (dynamic_pointer_cast<BigIntValue>(v)) return v;
    if (dynamic_pointer_cast<RingElemValue>(v)) return v;
    throw WrongTypeException(ListValue::type->name + " or " +
                             RingElemValue::type->name + " or " +
                             BigIntValue::type->name,
                             v->getType()->name, ARG(0).exp);
  }
  bool isBigInt0;
  intrusive_ptr<RightValue> v0 = runtimeEnv->evalArgAsT1orT2<BigIntValue, RingElemValue>(ARG(0), isBigInt0);
  bool isBigInt1;
  intrusive_ptr<RightValue> v1 = runtimeEnv->evalArgAsT1orT2<BigIntValue, RingElemValue>(ARG(1), isBigInt1);
  if (isBigInt0 && isBigInt1)
  {
    return Value::from(gcd(PtrCastBigInt(v0), PtrCastBigInt(v1)));
  }
  if (isBigInt0)
  {
    ring R = owner(PtrCastRingElem(v1));
    RingElem tmp(R, PtrCastBigInt(v0));
    return Value::from(gcd(tmp, PtrCastRingElem(v1)));
  }
  if (isBigInt1)
  {
    ring R = owner(PtrCastRingElem(v0));
    RingElem tmp(R, PtrCastBigInt(v1));
    return Value::from(gcd(PtrCastRingElem(v0), tmp));
  }
  return Value::from(gcd(PtrCastRingElem(v0), PtrCastRingElem(v1)));
//   bool isBigInt;
//   intrusive_ptr<RightValue> v0 = runtimeEnv->evalArgAsT1orT2<BigIntValue, RingElemValue>(ARG(0), isBigInt);
//   if (isBigInt)
//   {
//     intrusive_ptr<BigIntValue> v1 = runtimeEnv->evalArgAs<BigIntValue>(ARG(1));
//     return new BigIntValue(gcd(intrusive_ptr_cast<BigIntValue>(v0)->theInteger, v1->theInteger));
//   }
//   intrusive_ptr<RingElemValue> v1 = runtimeEnv->evalArgAs<RingElemValue>(ARG(1));
//   return new RingElemValue(gcd(intrusive_ptr_cast<RingElemValue>(v0)->theRingElem, v1->theRingElem));
}

DECLARE_ARITYCHECK_FUNCTION(lcm) { return (1<=nArg) && (nArg<=2); }
DECLARE_BUILTIN_FUNCTION(lcm)
{
  invocationExpression->checkNumberOfArgs(1,2);
  if (invocationExpression->args.size()==1)
  {
    intrusive_ptr<RightValue> v = runtimeEnv->evalArgAs<RightValue>(ARG(0));
    if (intrusive_ptr<ListValue> l = dynamic_pointer_cast<ListValue>(v)) {
      if (l->size() == 0) return BigIntValue::one;
      vector<BigInt> v1;
      if (IsVectorBigInt(v1, l)) return Value::from(lcm_forC5(v1));
      return Value::from(lcm_forC5(runtimeEnv->evalArgAsRingElemList(ARG(0))));
    }
    if (dynamic_pointer_cast<BigIntValue>(v)) return v;
    if (dynamic_pointer_cast<RingElemValue>(v)) return v;
    throw WrongTypeException(ListValue::type->name + " or " +
                             RingElemValue::type->name + " or " +
                             BigIntValue::type->name,
                             v->getType()->name, ARG(0).exp);
  }
  bool isBigInt0;
  intrusive_ptr<RightValue> v0 = runtimeEnv->evalArgAsT1orT2<BigIntValue, RingElemValue>(ARG(0), isBigInt0);
  bool isBigInt1;
  intrusive_ptr<RightValue> v1 = runtimeEnv->evalArgAsT1orT2<BigIntValue, RingElemValue>(ARG(1), isBigInt1);
  if (isBigInt0 && isBigInt1)
  {
    return Value::from(lcm(PtrCastBigInt(v0), PtrCastBigInt(v1)));
  }
  if (isBigInt0)
  {
    ring R = owner(PtrCastRingElem(v1));
    RingElem tmp(R, PtrCastBigInt(v0));
    return Value::from(lcm(tmp, PtrCastRingElem(v1)));
  }
  if (isBigInt1)
  {
    ring R = owner(PtrCastRingElem(v0));
    RingElem tmp(R, PtrCastBigInt(v1));
    return Value::from(lcm(PtrCastRingElem(v0), tmp));
  }
  return Value::from(lcm(PtrCastRingElem(v0), PtrCastRingElem(v1)));
}


DECLARE_STD_BUILTIN_FUNCTION(num, 1) { // AMB (changed)
	const Argument &arg = ARG(0);
  int which;
	intrusive_ptr<RightValue> v = runtimeEnv->evalArgAsT1orT2orT3<BigIntValue, BigRatValue, RingElemValue>(arg, which);
  switch (which) {
  case 1: return v;
  case 2: return Value::from(num(PtrCastBigRat(v)));
  default:return Value::from(num(PtrCastRingElem(v)));
  }
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(den, 1) {
	const Argument &arg = ARG(0);
  int which;
	intrusive_ptr<RightValue> v = runtimeEnv->evalArgAsT1orT2orT3<BigIntValue, BigRatValue, RingElemValue>(arg, which);
  switch (which) {
  case 1: return Value::from(BigInt(1));
  case 2: return Value::from(den(PtrCastBigRat(v)));
  default:return Value::from(den(PtrCastRingElem(v)));
  }
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(SimplestRATBetween, 2) { //JAA 2012-12-11
	const Argument &arg0 = ARG(0);
	const Argument &arg1 = ARG(1);
        BigRat A, B;
        bool ArgIsInt;
	intrusive_ptr<RightValue> v = runtimeEnv->evalArgAsT1orT2<BigIntValue, BigRatValue>(arg0, ArgIsInt);
        if (ArgIsInt)
          A = PtrCastBigInt(v);
        else
          A = PtrCastBigRat(v);

	v = runtimeEnv->evalArgAsT1orT2<BigIntValue, BigRatValue>(arg1, ArgIsInt);
        if (ArgIsInt)
          B = PtrCastBigInt(v);
        else
          B = PtrCastBigRat(v);

        return new BigRatValue(SimplestBigRatBetween(A, B));
}
END_STD_BUILTIN_FUNCTION


DECLARE_STD_BUILTIN_FUNCTION(sprint, 1) {
	const Argument &arg = ARG(0);
	intrusive_ptr<OutputStringStreamValue> outs(new OutputStringStreamValue);
	outs->print(runtimeEnv, runtimeEnv->evalArgAs<RightValue>(arg), arg.exp);
	return outs->close(runtimeEnv, invocationExpression);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(GetErrMesg, 1) {
	const Argument &arg = ARG(0);
	intrusive_ptr<ErrorValue> err = runtimeEnv->evalArgAs<ErrorValue>(arg);
	return new StringValue(err->message);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(close, 1) {
	const Argument &arg = ARG(0);
	return runtimeEnv->evalArgAs<OutputStreamValue>(arg)->close(runtimeEnv, invocationExpression);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(OpenOString, 0) {
	(void)runtimeEnv; // to avoid a compilation warning
	return new OutputStringStreamValue();
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(GetEnv, 1) {
	const Argument &arg = ARG(0);
	intrusive_ptr<StringValue> varName = runtimeEnv->evalArgAs<StringValue>(arg);
	char *value = ::getenv(varName->theString.c_str());
	if (!value)
		return StringValue::empty;
	return new StringValue(string(value));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(OpenOFile, 1) {
	const Argument &arg = ARG(0);
	intrusive_ptr<StringValue> filename = runtimeEnv->evalArgAs<StringValue>(arg);
	return new OutputFileStreamValue(filename, arg.exp);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(ContentsOfFile, 1) {
	const Argument &arg = ARG(0);
	intrusive_ptr<StringValue> filename = runtimeEnv->evalArgAs<StringValue>(arg);
	filtering_istream input;
	input.push(newline_filter(newline::posix));
	const file_source fs(filename->theString.c_str());
	if (!fs.is_open())
		throw RuntimeException("Cannot open file \""+filename->theString+"\" for reading.", invocationExpression);
	input.push(fs);
	string wholeFile;
	for(;;) {
		string line;
		getline(input, line);
		if (input.bad())
			throw RuntimeException("Cannot read from \""+filename->theString+"\".", invocationExpression);
		wholeFile += line;
		if (input.eof())
			break;
		wholeFile += '\n';
	}
	return new StringValue(wholeFile);
}
END_STD_BUILTIN_FUNCTION

const string invalidTagError("The tag must be a valid identifier");

DECLARE_STD_BUILTIN_FUNCTION(tagged, 2) {
	const Argument &argValue = ARG(0);
	const Argument &argTag = ARG(1);
	intrusive_ptr<RightValue> v = runtimeEnv->evalArgAs<RightValue>(argValue);
	const string tag = runtimeEnv->evalArgAs<StringValue>(argTag)->theString;
	// TODO: must be fixed
	//if (!LexerNS::Lexer::isValidIdentifier(tag))
	//	throw RuntimeException(invalidTagError, argTag.exp);
	return new TaggedValue(v->untagged(), invocationExpression->packageName+"."+tag);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(untagged, 1) {
	return runtimeEnv->evalArgAs<RightValue>(ARG(0))->untagged();
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(tag, 1) {
	const Argument &argValue = ARG(0);
	intrusive_ptr<RightValue> v = runtimeEnv->evalArgAs<RightValue>(argValue);
	if (intrusive_ptr<TaggedValue> taggedValue = dynamic_pointer_cast<TaggedValue>(v)) {
		const string::size_type l = invocationExpression->packageName.length()+1;
		if (taggedValue->tag.substr(0, l)==invocationExpression->packageName+".")
			return new StringValue(taggedValue->tag.substr(l, string::npos));
		return new StringValue(taggedValue->tag);
	}
	return StringValue::empty;
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(type, 1) {
	return runtimeEnv->evalArgAs<RightValue>(ARG(0))->getType();
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(PkgName, 0) {
	(void)runtimeEnv; // to avoid a compilation warning
	return new StringValue(invocationExpression->packageName);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(/* uppercase because it's a backward-compatibility hack */ Functions, 1) {
	(void)runtimeEnv; // to avoid a compilation warning
	throw RuntimeException("Use the statement Describe to obtain package information", invocationExpression);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(TAGGED, 1) {
	const Argument &arg = ARG(0);
	const string tag = runtimeEnv->evalArgAs<StringValue>(arg)->theString;
	if (!LexerNS::Lexer::isValidIdentifier(tag))
		throw RuntimeException(invalidTagError, arg.exp);
	return TypeValue::tagType(tag);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(types, 0) {
	(void)runtimeEnv; // to avoid a compilation warning
	throw RuntimeException("Cannot return a list of all types because they're infinite (CoCoA 4 lies about this, I don't). You might want to use CurrentTypes() to get a list of currently instantiated types", invocationExpression);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(CurrentTypes, 0) {
	return runtimeEnv->currentTypes();
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(error, 1) {
	const Argument &arg = ARG(0);
	const string message = runtimeEnv->evalArgAs<StringValue>(arg)->theString;
	throw RuntimeException(message, invocationExpression);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(Error, 1) {
	const Argument &arg = ARG(0);
	const string message = runtimeEnv->evalArgAs<StringValue>(arg)->theString;
	throw RuntimeException(message, invocationExpression);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(ascii, 1) {
	const Argument &arg = ARG(0);
	intrusive_ptr<RightValue> v = runtimeEnv->evalArgAs<RightValue>(arg);
	if (intrusive_ptr<StringValue> sv = dynamic_pointer_cast<StringValue>(v)) {
		intrusive_ptr<ListValue> l(new ListValue());
		const string s(sv->theString);
		BOOST_FOREACH(const char c, s)
			l->addValue(new BigIntValue(static_cast<int>(c)));
		return l;
	}
	if (intrusive_ptr<BigIntValue> N = dynamic_pointer_cast<BigIntValue>(v)) {
		long l;
		if (!IsConvertible(l, N->theInteger) || l<0 || l>255)
			throw RuntimeException("When the argument is an integer, it must be in the range [0..255]", arg.exp);
		return new StringValue(string(1, static_cast<char>(l)));
	}
	if (intrusive_ptr<ListValue> list = dynamic_pointer_cast<ListValue>(v)) {
		string retValue;
		ListValue::ContainerType::size_type size = list->size();
		for(ListValue::ContainerType::size_type a=0; a<size; ++a) {
			intrusive_ptr<BigIntValue> N = dynamic_pointer_cast<BigIntValue>(list->getValue(a));
			long l;
			if (!N || !IsConvertible(l, N->theInteger) || l<0 || l>255)
				throw RuntimeException("When the argument is a list, it must be a list of integers only in the range [0..255]", arg.exp);
			retValue += static_cast<char>(l);
		}
		return Value::from(retValue);
	}
	throw WrongTypeException(ListValue::type->name+", "+StringValue::type->name + " or "+BigIntValue::type->name, v->getType()->name, arg.exp);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(ILogBase, 2) { // AMB
	const Argument &a0 = ARG(0);
  intrusive_ptr<BigIntValue> base = runtimeEnv->evalArgAs<BigIntValue>(ARG(1));
	bool isBigInt;
	intrusive_ptr<RightValue> v0 = runtimeEnv->evalArgAsT1orT2<BigIntValue, BigRatValue>(a0, isBigInt);
	if (isBigInt)
    return Value::from(ILogBase(runtimeEnv->evalArgAs<BigIntValue>(a0)->theInteger, base->theInteger));
    return Value::from(ILogBase(runtimeEnv->evalArgAs<BigRatValue>(a0)->theRational, base->theInteger));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(log, 1) { // AMB
  intrusive_ptr<RingElemValue> f = runtimeEnv->evalArgAs<RingElemValue>(ARG(0));
  vector<BigInt> expv;
  if (NumIndets(AsPolyRing(owner(f->theRingElem)))==1)
    expv.push_back(BigInt(deg(f->theRingElem)));
  else 
    BigExponents(expv, LPP(f->theRingElem));
	return new ListValue(expv);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(CoeffRing, 1) { // AMB
	const Argument &arg = ARG(0);
	return new RingValue(CoeffRing(AsPolyRing((runtimeEnv->evalArgAs<RingValue>(arg))->theRing)));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(DefiningIdeal, 1) { // AMB
	const Argument &arg = ARG(0);
	return Value::from(DefiningIdeal(AsQuotientRing((runtimeEnv->evalArgAs<RingValue>(arg))->theRing)));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(wdeg, 1) { // AMB
	bool isFirst;
  intrusive_ptr<const RightValue> x = runtimeEnv->evalArgAsT1orT2<RingElemValue, ModuleElemValue>(ARG(0), isFirst);
	if (isFirst)
    //    return Value::from(wdeg_forC5(intrusive_ptr_cast<RingElemValue>(x)->theRingElem));
    return Value::from(wdeg_forC5(PtrCastRingElem(x)));
  return Value::from(wdeg_forC5(PtrCastModuleElem(x)));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(IsHomog, 1) { // AMB
	int which;
	intrusive_ptr<const RightValue> v = runtimeEnv->evalArgAsT1orT2orT3orT4<RingElemValue, ModuleElemValue, IdealValue, ListValue>(ARG(0), which);
  switch (which) {
  case 1: return Value::from(IsHomog(PtrCastRingElem(v)));
  case 2: return Value::from(IsHomog(PtrCastModuleElem(v)));
  case 3: return Value::from(IsHomog(PtrCastIdeal(v)));
  default:
    if (const intrusive_ptr<const ListValue> l = intrusive_ptr_cast<const ListValue>(v))
    {
      ListValue::ContainerType::size_type size = l->size();
      for(ListValue::ContainerType::size_type a=1; a<size; ++a)
      {
        if (const boost::intrusive_ptr<const RingElemValue> f = boost::dynamic_pointer_cast<const RingElemValue>(l->getValue(a)))
          if (!IsHomog(f->theRingElem)) return BoolValue::fromBool(false);
      }
      return BoolValue::fromBool(true);
    }
  }
  return BoolValue::fromBool(true); // just to keep the compiler quiet
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(LPP, 1) { // AMB
	bool isFirst;
	intrusive_ptr<RightValue> v = runtimeEnv->evalArgAsT1orT2<RingElemValue, ModuleElemValue>(ARG(0), isFirst);
  if (isFirst)
    return Value::from(LPP_forC5(PtrCastRingElem(v)));
  return Value::from(LPP_forC5(PtrCastModuleElem(v)));
}
END_STD_BUILTIN_FUNCTION


DECLARE_STD_BUILTIN_FUNCTION(LT, 1) { // AMB
	int which;
	intrusive_ptr<RightValue> v = runtimeEnv->evalArgAsT1orT2orT3<RingElemValue, ModuleElemValue, IdealValue>(ARG(0), which);
  switch (which) {
  case 1: return Value::from(LT_forC5(PtrCastRingElem(v)));
  case 2: return Value::from(LT_forC5(PtrCastModuleElem(v)));
  default:return Value::from(LT(PtrCastIdeal(v)));
  }
}
END_STD_BUILTIN_FUNCTION


DECLARE_STD_BUILTIN_FUNCTION(QuotientBasis, 1) { // JAA
	intrusive_ptr<RightValue> v = runtimeEnv->evalArgAs<RightValue>(ARG(0));
	if (const intrusive_ptr<const IdealValue> I = dynamic_pointer_cast<const IdealValue>(v))
    return Value::from(QuotientBasis_forC5(I->theIdeal));
//   if (const intrusive_ptr<const ListValue> l = dynamic_pointer_cast<const ListValue>(v))
//   {
//     ListValue::ContainerType::size_type size = l->size();
//     for(ListValue::ContainerType::size_type a=1; a<size; ++a)
//     {
//       if (const boost::intrusive_ptr<const RingElemValue> f = boost::dynamic_pointer_cast<const RingElemValue>(l->getValue(a)))
//         if (!IsHomog(f->theRingElem)) return BoolValue::fromBool(false);
//     }
//     return BoolValue::fromBool(true);
//   }
	throw WrongTypeException(IdealValue::type->name,
                           v->getType()->name, ARG(0).exp);
  return BoolValue::fromBool(true); // just to keep the compiler quiet
}
END_STD_BUILTIN_FUNCTION



DECLARE_STD_BUILTIN_FUNCTION(RingOf, 1) { // AMB
	intrusive_ptr<RightValue> v = runtimeEnv->evalArgAs<RightValue>(ARG(0));
	if (const intrusive_ptr<RingElemValue> x = dynamic_pointer_cast<RingElemValue>(v))
    return Value::from(owner(x->theRingElem));
	if (const intrusive_ptr<IdealValue> x = dynamic_pointer_cast<IdealValue>(v))
    return Value::from(AmbientRing(x->theIdeal));
	if (const intrusive_ptr<MatrixValue> x = dynamic_pointer_cast<MatrixValue>(v))
    return Value::from(BaseRing(x->theMatrix));
	if (const intrusive_ptr<ModuleValue> x = dynamic_pointer_cast<ModuleValue>(v))
    return Value::from(BaseRing(x->theModule));
	if (const intrusive_ptr<ModuleElemValue> x = dynamic_pointer_cast<ModuleElemValue>(v))
    return Value::from(BaseRing(owner(x->theModuleElem)));
//   if (const intrusive_ptr<RingValue> R = dynamic_pointer_cast<RingValue>(v))
//   {
//     if (IsFractionField(R->theRing))
//       return Value::from(BaseRing(AsFractionField(R->theRing)));
//     if (IsQuotientRing(R->theRing))
//       return Value::from(BaseRing(AsQuotientRing(R->theRing)));
//   }
	throw WrongTypeException(RingElemValue::type->name + " or " +
                           ModuleElemValue::type->name + " or " +
                           MatrixValue::type->name,
                           IdealValue::type->name + " or " +
                           ModuleValue::type->name + " or " +
                           v->getType()->name, ARG(0).exp);
  return Value::from(RingZZ()); // just to keep the compiler quiet
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(ModuleOf, 1) { // AMB
	intrusive_ptr<RightValue> v = runtimeEnv->evalArgAs<RightValue>(ARG(0));
	if (const intrusive_ptr<ModuleElemValue> x = dynamic_pointer_cast<ModuleElemValue>(v))
    return Value::from(owner(x->theModuleElem));
	throw WrongTypeException(ModuleElemValue::type->name + " or " +
                           ModuleValue::type->name,
                           v->getType()->name, ARG(0).exp);
  return Value::from(NewFreeModule(RingZZ(),1)); // just to keep the compiler quiet
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(BaseRing, 1) { // AMB
	intrusive_ptr<RightValue> v = runtimeEnv->evalArgAs<RightValue>(ARG(0));
	if (const intrusive_ptr<const MatrixValue> x = dynamic_pointer_cast<const MatrixValue>(v))
    return Value::from(BaseRing(x->theMatrix));
	if (const intrusive_ptr<const ModuleValue> x = dynamic_pointer_cast<const ModuleValue>(v))
    return Value::from(BaseRing(x->theModule));
  if (const intrusive_ptr<RingValue> x = dynamic_pointer_cast<RingValue>(v))
  {
    if (IsFractionField(x->theRing))
      return Value::from(BaseRing(AsFractionField(x->theRing)));
    if (IsQuotientRing(x->theRing))
      return Value::from(BaseRing(AsQuotientRing(x->theRing)));
  }
	throw WrongTypeException(MatrixValue::type->name + " or " +
                           ModuleValue::type->name + " or " +
                           RingValue::type->name,
                           v->getType()->name, ARG(0).exp);
  return new RingValue(RingZZ()); // just to keep the compiler quiet
}
END_STD_BUILTIN_FUNCTION

//---- RINGHOM -------------------------------------------------------

DECLARE_STD_BUILTIN_FUNCTION(CoeffEmbeddingHom, 1) { // AMB
  PolyRing P = AsPolyRing((runtimeEnv->evalArgAs<RingValue>(ARG(0)))->theRing);
	return new RingHomValue(CoeffEmbeddingHom(P));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(EmbeddingHom, 1) { // AMB
  intrusive_ptr<RingValue> R = runtimeEnv->evalArgAs<RingValue>(ARG(0));
	return new RingHomValue(EmbeddingHom(AsFractionField(R->theRing)));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(InducedHom, 2) { // AMB
  intrusive_ptr<RingValue> R = runtimeEnv->evalArgAs<RingValue>(ARG(0));
  intrusive_ptr<RingHomValue> phi = runtimeEnv->evalArgAs<RingHomValue>(ARG(1));
  if (IsFractionField(R->theRing))
    return Value::from(InducedHom(AsFractionField(R->theRing),phi->theRingHom));
  if (IsQuotientRing(R->theRing))
    return Value::from(InducedHom(AsQuotientRing(R->theRing), phi->theRingHom));
  throw RuntimeException("FractionField or QuotientRing expected", ARG(0).exp);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(PolyAlgebraHom, 3) { // AMB
  intrusive_ptr<RingValue> P = runtimeEnv->evalArgAs<RingValue>(ARG(0));
  intrusive_ptr<RingValue> R = runtimeEnv->evalArgAs<RingValue>(ARG(1));
  vector<RingElem> IndetImages = runtimeEnv->evalArgAsRingElemList(ARG(2), R->theRing);
	return Value::from(PolyAlgebraHom(AsPolyRing(P->theRing), R->theRing, IndetImages));
}
END_STD_BUILTIN_FUNCTION 

DECLARE_STD_BUILTIN_FUNCTION(PolyRingHom, 4) { // AMB
  intrusive_ptr<RingValue> P = runtimeEnv->evalArgAs<RingValue>(ARG(0));
  intrusive_ptr<RingValue> R = runtimeEnv->evalArgAs<RingValue>(ARG(1));
  intrusive_ptr<RingHomValue> CoeffHom = runtimeEnv->evalArgAs<RingHomValue>(ARG(2));
  vector<RingElem> IndetImages = runtimeEnv->evalArgAsRingElemList(ARG(3), R->theRing);
	return new RingHomValue(PolyRingHom(AsPolyRing(P->theRing), R->theRing, CoeffHom->theRingHom, IndetImages));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(apply, 2) {
	intrusive_ptr<RingHomValue> phi = runtimeEnv->evalArgAs<RingHomValue>(ARG(0));
	intrusive_ptr<RightValue> v = runtimeEnv->evalArgAs<RightValue>(ARG(1));
	if (intrusive_ptr<RingElemValue> x = dynamic_pointer_cast<RingElemValue>(v))
		return Value::from(apply(phi->theRingHom, x->theRingElem));
	if (intrusive_ptr<MatrixValue> x = dynamic_pointer_cast<MatrixValue>(v))
		return Value::from(apply(phi->theRingHom, x->theMatrix));
	if (intrusive_ptr<ListValue> x = dynamic_pointer_cast<ListValue>(v))
  {
    vector<RingElem> v = runtimeEnv->evalArgAsRingElemList(ARG(1), domain(phi->theRingHom));
		return Value::from(apply(phi->theRingHom, v));
  }
	throw WrongTypeException(RingElemValue::type->name + " or " +
                           ListValue::type->name + " or " +
                           MatrixValue::type->name,
                           v->getType()->name, ARG(1).exp);
}
END_STD_BUILTIN_FUNCTION

// DECLARE_STD_BUILTIN_FUNCTION(ord, 1) { // AMB
// 	return Value::from(ordering(PPM(AsSparsePolyRing((runtimeEnv->evalArgAs<RingValue>(ARG(0)))->theRing))));
// }
// END_STD_BUILTIN_FUNCTION

// variable number of args
DECLARE_ARITYCHECK_FUNCTION(indets) { return (1<=nArg) && (nArg<=2); }
DECLARE_BUILTIN_FUNCTION(indets) { // AMB+JAA
  invocationExpression->checkNumberOfArgs(1,2);
  if (invocationExpression->args.size()==1)
    return Value::from(indets(AsPolyRing((runtimeEnv->evalArgAs<RingValue>(ARG(0)))->theRing)));
  return Value::from(indets(AsPolyRing((runtimeEnv->evalArgAs<RingValue>(ARG(0)))->theRing),
                            runtimeEnv->evalArgAs<StringValue>(ARG(1))->theString));
}


DECLARE_STD_BUILTIN_FUNCTION(NumIndets, 1) { // AMB
	return Value::from(NumIndets(AsPolyRing((runtimeEnv->evalArgAs<RingValue>(ARG(0)))->theRing)));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(indet, 2) { // AMB
	intrusive_ptr<RingValue> a = runtimeEnv->evalArgAs<RingValue>(ARG(0));
	intrusive_ptr<BigIntValue> b = runtimeEnv->evalArgAs<BigIntValue>(ARG(1));
  long n;
  if (!IsConvertible(n, b->theInteger))
    throw RuntimeException("invalid indet index", ARG(1).exp);
	return Value::from(indet(AsPolyRing(a->theRing), n-1));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(IndetIndex, 1) { // AMB
  long i;
  if (!IsIndet(i, runtimeEnv->evalArgAs<RingElemValue>(ARG(0))->theRingElem))
    throw RuntimeException("not an indet", ARG(0).exp);
  return Value::from(i+1);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(IndetName, 1) { // AMB
	intrusive_ptr<RingElemValue> a = runtimeEnv->evalArgAs<RingElemValue>(ARG(0));
  long i;
  if (!IsIndet(i,a->theRingElem))
    throw RuntimeException("not an indet", ARG(0).exp);
  symbol s=IndetSymbol(AsPolyRing(owner(a->theRingElem)), i);
	return Value::from(head(s));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(IndetSubscripts, 1) { // AMB
	intrusive_ptr<RingElemValue> a = runtimeEnv->evalArgAs<RingElemValue>(ARG(0));
  long i;
  if (!IsIndet(i,a->theRingElem))
    throw RuntimeException("not an indet", ARG(0).exp);
  symbol s=IndetSymbol(AsPolyRing(owner(a->theRingElem)), i);
  vector<long> iss;
  for (long n=0; n<NumSubscripts(s); ++n) iss.push_back(subscript(s,n));
	return Value::from(iss);
}
END_STD_BUILTIN_FUNCTION


DECLARE_STD_BUILTIN_FUNCTION(ContentWRT, 2) { // AMB
	intrusive_ptr<RingElemValue> f = runtimeEnv->evalArgAs<RingElemValue>(ARG(0));
	intrusive_ptr<RightValue> idt = runtimeEnv->evalArgAs<RightValue>(ARG(1));
  if (intrusive_ptr<RingElemValue> x = dynamic_pointer_cast<RingElemValue>(idt))
    return Value::from(ContentWRT(f->theRingElem, x->theRingElem));
  if (intrusive_ptr<ListValue> l = dynamic_pointer_cast<ListValue>(idt))
  {
//     vector<long> v1 = VectorLongDecr1(runtimeEnv->evalArgAsListOf<BigIntValue>(ARG(1)), ERR::BadIndetIndex, "ContentWRT");
//     return Value::from(ContentWRT(f->theRingElem, v1));
    vector<RingElem> v1 = runtimeEnv->evalArgAsRingElemList(ARG(1), owner(f->theRingElem));
    return Value::from(ContentWRT_forC5(f->theRingElem, v1));
  }
  throw WrongTypeException(ListValue::type->name + " or " +
                           RingElemValue::type->name,
                           idt->getType()->name, ARG(1).exp);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(CoefficientsWRT, 2) { // AMB
	intrusive_ptr<RingElemValue> f = runtimeEnv->evalArgAs<RingElemValue>(ARG(0));
	intrusive_ptr<RightValue> idt = runtimeEnv->evalArgAs<RightValue>(ARG(1));
  std::vector<CoeffPP> CoeffWRT;
  if (intrusive_ptr<RingElemValue> x = dynamic_pointer_cast<RingElemValue>(idt))
    CoeffWRT = CoefficientsWRT(f->theRingElem, x->theRingElem);
  else
  if (intrusive_ptr<ListValue> l = dynamic_pointer_cast<ListValue>(idt))
  {
//     vector<long> v1 = VectorLongDecr1(runtimeEnv->evalArgAsListOf<BigIntValue>(ARG(1)), ERR::BadIndetIndex, "ContentWRT");
//     CoeffWRT = CoefficientsWRT(f->theRingElem, v1);
    vector<RingElem> v1 = runtimeEnv->evalArgAsRingElemList(ARG(1), owner(f->theRingElem));
    CoeffWRT = CoefficientsWRT_forC5(f->theRingElem, v1);
  }
  else
    throw WrongTypeException(ListValue::type->name + " or " +
                             RingElemValue::type->name,
                             idt->getType()->name, ARG(1).exp);
  // create return value: list of ...
  SparsePolyRing P = AsSparsePolyRing(owner(f->theRingElem));
	intrusive_ptr<ListValue> returnValue(new ListValue);
  BOOST_FOREACH(const CoeffPP cpp, CoeffWRT)
  {
    // create return value: ... record
    intrusive_ptr<RecordValue> cpp5(new RecordValue);
    cpp5->setField("Coeff", Value::from(cpp.myCoeff));
    cpp5->setField("PP",    Value::from(monomial(P,1,cpp.myPP)));
    returnValue->addValue(cpp5);
	}
  return returnValue;
}
END_STD_BUILTIN_FUNCTION



//---- MAT -------------------------------------------------------

DECLARE_STD_BUILTIN_FUNCTION(SetEntry, 4) { // AMB
	intrusive_ptr<MatrixValue> M = runtimeEnv->obtainUnshared<MatrixValue>(ARG(0));
	intrusive_ptr<BigIntValue> I = runtimeEnv->evalArgAs<BigIntValue>(ARG(1));
	intrusive_ptr<BigIntValue> J = runtimeEnv->evalArgAs<BigIntValue>(ARG(2));
  int which;
	intrusive_ptr<RightValue> v = runtimeEnv->evalArgAsT1orT2orT3<BigIntValue, BigRatValue, RingElemValue>(ARG(3), which);
  RingElem x(BaseRing(M->theMatrix));
	if (which==1) x = PtrCastBigInt(v);
	if (which==2) x = PtrCastBigRat(v);
	if (which==3) x = PtrCastRingElem(v);
	SetEntry_forC5(M->theMatrix, I->theInteger, J->theInteger, x);
	return VoidValue::theInstance;
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(SetRow, 3) { // AMB
	intrusive_ptr<MatrixValue> M = runtimeEnv->obtainUnshared<MatrixValue>(ARG(0));
	intrusive_ptr<BigIntValue> I = runtimeEnv->evalArgAs<BigIntValue>(ARG(1));
  vector<RingElem> v = runtimeEnv->evalArgAsRingElemList(ARG(2), BaseRing(M->theMatrix));
	SetRow_forC5(M->theMatrix, I->theInteger, v);
	return VoidValue::theInstance;
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(SetCol, 3) { // AMB
	intrusive_ptr<MatrixValue> M = runtimeEnv->obtainUnshared<MatrixValue>(ARG(0));
	intrusive_ptr<BigIntValue> I = runtimeEnv->evalArgAs<BigIntValue>(ARG(1));
  vector<RingElem> v = runtimeEnv->evalArgAsRingElemList(ARG(2), BaseRing(M->theMatrix));
	SetCol_forC5(M->theMatrix, I->theInteger, v);
	return VoidValue::theInstance;
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(SwapRows, 3) { // AMB
	intrusive_ptr<MatrixValue> M = runtimeEnv->obtainUnshared<MatrixValue>(ARG(0));
	intrusive_ptr<BigIntValue> I1 = runtimeEnv->evalArgAs<BigIntValue>(ARG(1));
	intrusive_ptr<BigIntValue> I2 = runtimeEnv->evalArgAs<BigIntValue>(ARG(2));
	SwapRows_forC5(M->theMatrix, I1->theInteger, I2->theInteger);
	return VoidValue::theInstance;
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(SwapCols, 3) { // AMB
	intrusive_ptr<MatrixValue> M = runtimeEnv->obtainUnshared<MatrixValue>(ARG(0));
	intrusive_ptr<BigIntValue> I1 = runtimeEnv->evalArgAs<BigIntValue>(ARG(1));
	intrusive_ptr<BigIntValue> I2 = runtimeEnv->evalArgAs<BigIntValue>(ARG(2));
	SwapCols_forC5(M->theMatrix, I1->theInteger, I2->theInteger);
	return VoidValue::theInstance;
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(AssignZero, 1) { // AMB
	intrusive_ptr<MatrixValue> M = runtimeEnv->obtainUnshared<MatrixValue>(ARG(0));
	AssignZero(M->theMatrix);
	return VoidValue::theInstance;
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(IsZeroRow, 2) { // AMB
	intrusive_ptr<MatrixValue> M = runtimeEnv->evalArgAs<MatrixValue>(ARG(0));
	intrusive_ptr<BigIntValue> N = runtimeEnv->evalArgAs<BigIntValue>(ARG(1));
  long n;
  if (!IsConvertible(n, N->theInteger))
    throw RuntimeException("invalid row index", ARG(1).exp);
  return BoolValue::fromBool(IsZeroRow(M->theMatrix, n-1));
	//return new BoolValue(IsZeroRow(M->theMatrix, n));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(IsZeroCol, 2) { // AMB
	intrusive_ptr<MatrixValue> M = runtimeEnv->evalArgAs<MatrixValue>(ARG(0));
	intrusive_ptr<BigIntValue> N = runtimeEnv->evalArgAs<BigIntValue>(ARG(1));
  long n;
  if (!IsConvertible(n, N->theInteger))
    throw RuntimeException("invalid col index", ARG(1).exp);
  return BoolValue::fromBool(IsZeroCol(M->theMatrix, n-1));
	//return new BoolValue(IsZeroCol(M->theMatrix, n));
}
END_STD_BUILTIN_FUNCTION

// variable number of args
DECLARE_ARITYCHECK_FUNCTION(IsPositiveGrading) { return (1<=nArg) && (nArg<=2); }
DECLARE_BUILTIN_FUNCTION(IsPositiveGrading) { // AMB variable number of args
	invocationExpression->checkNumberOfArgs(1,2);
	intrusive_ptr<MatrixValue> M = runtimeEnv->evalArgAs<MatrixValue>(ARG(0));
  if (invocationExpression->args.size()==1)
    return BoolValue::fromBool(IsPositiveGrading(M->theMatrix)); // 1 arg: done
  intrusive_ptr<BigIntValue> N = runtimeEnv->evalArgAs<BigIntValue>(ARG(1));
  long GrDim = NumRows(M->theMatrix);
  if (!IsConvertible(GrDim, N->theInteger))
    throw RuntimeException("invalid row index", ARG(1).exp);
  return BoolValue::fromBool(IsPositiveGrading(M->theMatrix, GrDim));
}
//END_STD_BUILTIN_FUNCTION // no: variable number of args

DECLARE_STD_BUILTIN_FUNCTION(ElimMat, 2) {
	bool isFirst;
  intrusive_ptr<RightValue> x = runtimeEnv->evalArgAsT1orT2<BigIntValue, MatrixValue>(ARG(0), isFirst);
  vector<BigInt> elim = runtimeEnv->evalArgAsListOf<BigIntValue>(ARG(1));
	if (isFirst)
    return Value::from(ElimMat_forC5(PtrCastBigInt(x), elim));
  return Value::from(ElimMat_forC5(PtrCastMatrix(x), elim));
}
END_STD_BUILTIN_FUNCTION


DECLARE_STD_BUILTIN_FUNCTION(submat, 3) { // AMB
	intrusive_ptr<MatrixValue> M = runtimeEnv->evalArgAs<MatrixValue>(ARG(0));
  vector<long> rows = VectorLongDecr1(runtimeEnv->evalArgAsListOf<BigIntValue>(ARG(1)), ERR::BadRowIndex, "submat");
  vector<long> cols = VectorLongDecr1(runtimeEnv->evalArgAsListOf<BigIntValue>(ARG(2)), ERR::BadColIndex, "submat");
	return new MatrixValue(NewDenseMat(submat(M->theMatrix, rows, cols)));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(factor, 1) { // AMB
	intrusive_ptr<RingElemValue> f = runtimeEnv->evalArgAs<RingElemValue>(ARG(0));
  factorization<RingElem> facs = factor(f->theRingElem);
  intrusive_ptr<RecordValue> facsC5(new RecordValue);
  facsC5->setField("Factors", Value::from(facs.myFactors));
  facsC5->setField("Multiplicities", Value::from(facs.myMultiplicities));
  facsC5->setField("RemainingFactor", Value::from(facs.myRemainingFactor));
  return facsC5;
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(SmoothFactor, 2) { // AMB
	intrusive_ptr<BigIntValue> N = runtimeEnv->evalArgAs<BigIntValue>(ARG(0));
	intrusive_ptr<BigIntValue> TrialLimit = runtimeEnv->evalArgAs<BigIntValue>(ARG(1));
  factorization<BigInt> facs = SmoothFactor_forC5(N->theInteger, TrialLimit->theInteger);
  intrusive_ptr<RecordValue> facsC5(new RecordValue);
  facsC5->setField("Factors", Value::from(facs.myFactors));
  facsC5->setField("Multiplicities", Value::from(facs.myMultiplicities));
  facsC5->setField("RemainingFactor", new BigIntValue(facs.myRemainingFactor));
  return facsC5;
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(ContentFreeFactor, 1) { // AMB
	intrusive_ptr<RingElemValue> f = runtimeEnv->evalArgAs<RingElemValue>(ARG(0));
  factorization<RingElem> facs = ContentFreeFactor(f->theRingElem);
  intrusive_ptr<RecordValue> facsC5(new RecordValue);
  facsC5->setField("Factors", Value::from(facs.myFactors));
  facsC5->setField("Multiplicities", Value::from(facs.myMultiplicities));
  facsC5->setField("RemainingFactor", Value::from(facs.myRemainingFactor));
  return facsC5;
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(NR, 2) { // AMB
	intrusive_ptr<RingElemValue> f = runtimeEnv->evalArgAs<RingElemValue>(ARG(0));
	vector<RingElem> v = runtimeEnv->evalArgAsListOf<RingElemValue>(ARG(1));
	return Value::from(NR(f->theRingElem,v));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(BinSequentialToric, 2) { // AMB
  intrusive_ptr<IdealValue> I = runtimeEnv->evalArgAs<IdealValue>(ARG(0));
  vector<BigInt> indices = runtimeEnv->evalArgAsListOf<BigIntValue>(ARG(1));
	return Value::from(SequentialToric_C(I->theIdeal, VectorLong(indices, "BinSequentialToric")));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(MatSequentialToric, 2) { // AMB
  intrusive_ptr<RingValue> R = runtimeEnv->evalArgAs<RingValue>(ARG(0));
  intrusive_ptr<MatrixValue> M = runtimeEnv->evalArgAs<MatrixValue>(ARG(1));
	return Value::from(SequentialToric_C(AsSparsePolyRing(R->theRing), M->theMatrix));
}
END_STD_BUILTIN_FUNCTION

// variable number of args
DECLARE_ARITYCHECK_FUNCTION(ideal) { return 1<=nArg; }
DECLARE_BUILTIN_FUNCTION(ideal) {
	const int nArgs = invocationExpression->args.size();
	if (nArgs==0)
		throw RuntimeException("Wrong number of arguments; found: 0, expecting: at least 1", invocationExpression);
	intrusive_ptr<RightValue> x = runtimeEnv->evalArgAs<RightValue>(ARG(0));
  if (invocationExpression->args.size()==2)
    if (intrusive_ptr<RingValue> R = dynamic_pointer_cast<RingValue>(x))
      return Value::from(ideal(R->theRing, runtimeEnv->evalArgAsRingElemList(ARG(1), R->theRing)));
	return Value::from(ideal(runtimeEnv->evalAllArgsAsListOf<RingElemValue>(invocationExpression)));
}

//---- MODULE -------------------------------------------------------

DECLARE_STD_BUILTIN_FUNCTION(ModuleElem, 2) { // AMB
  intrusive_ptr<ModuleValue> M = runtimeEnv->evalArgAs<ModuleValue>(ARG(0));
	vector<RingElem> v = runtimeEnv->evalArgAsRingElemList(ARG(1), BaseRing(M->theModule));
  return Value::from(NewFreeModuleElem(M->theModule, v));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(NewFreeModule, 2) { // AMB
  intrusive_ptr<RingValue> R = runtimeEnv->evalArgAs<RingValue>(ARG(0));
  intrusive_ptr<BigIntValue> N = runtimeEnv->evalArgAs<BigIntValue>(ARG(1));
  long n;
  if (!IsConvertible(n, N->theInteger) || n<0)
    throw RuntimeException("invalid length", ARG(1).exp);
	return Value::from(NewFreeModule(R->theRing, n));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(NewFreeModuleForSyz, 1) { // AMB
	int which;
	intrusive_ptr<RightValue> v = runtimeEnv->evalArgAsT1orT2orT3<IdealValue, ModuleValue, ListValue>(ARG(0), which);
  switch (which) {
  case 1: return Value::from(NewFreeModuleForSyz(gens(PtrCastIdeal(v))));
  case 2: return Value::from(NewFreeModuleForSyz(gens(AsFGModule(PtrCastModule(v)))));
  default:
    vector<RingElem> x = runtimeEnv->evalArgAsRingElemList(ARG(0));
    return Value::from(NewFreeModuleForSyz(x));
  }
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(submodule, 2) { // AMB
  intrusive_ptr<ModuleValue> F = runtimeEnv->evalArgAs<ModuleValue>(ARG(0));
  vector<ModuleElem> v1 = runtimeEnv->evalArgAsListOf<ModuleElemValue>(ARG(1));
  return Value::from(submodule(AsFGModule(F->theModule), v1));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(GensAsRows, 1) { // AMB
  intrusive_ptr<ModuleValue> F = runtimeEnv->evalArgAs<ModuleValue>(ARG(0));
  return Value::from(GensAsRows(AsFGModule(F->theModule)));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(GensAsCols, 1) { // AMB
  intrusive_ptr<ModuleValue> F = runtimeEnv->evalArgAs<ModuleValue>(ARG(0));
  return Value::from(GensAsCols(AsFGModule(F->theModule)));
}
END_STD_BUILTIN_FUNCTION


//---- IDEAL -------------------------------------------------------

// variable number of args
DECLARE_ARITYCHECK_FUNCTION(janet) { return 1<=nArg; }
DECLARE_BUILTIN_FUNCTION(janet) {
	const int nArgs = invocationExpression->args.size();
	if (nArgs==0)
		throw RuntimeException("Wrong number of arguments; found: 0, expecting: at least 1", invocationExpression);
	intrusive_ptr<RightValue> x = runtimeEnv->evalArgAs<RightValue>(ARG(0));
  if (invocationExpression->args.size()==2)
    if (intrusive_ptr<RingValue> R = dynamic_pointer_cast<RingValue>(x))
      return Value::from(ExtendedJanetBasis(runtimeEnv->evalArgAsRingElemList(ARG(1), R->theRing)));
	return Value::from(ExtendedJanetBasis(runtimeEnv->evalAllArgsAsListOf<RingElemValue>(invocationExpression)));
}

//---- RING -------------------------------------------------------

DECLARE_STD_BUILTIN_FUNCTION(NewRingTwinFloat, 1) { // AMB
	intrusive_ptr<BigIntValue> prec = runtimeEnv->evalArgAs<BigIntValue>(ARG(0));
  long d;
  if (!IsConvertible(d, prec->theInteger))
		throw RuntimeException("invalid precision", ARG(0).exp);
	return Value::from(NewRingTwinFloat(d));
}
END_STD_BUILTIN_FUNCTION


// variable number of args
DECLARE_ARITYCHECK_FUNCTION(NewPolyRing) { return (nArg==2) || (nArg==4); }
DECLARE_BUILTIN_FUNCTION(NewPolyRing) {
  //	invocationExpression->checkNumberOfArgs(2,4);
	const int nArgs = invocationExpression->args.size();
	if (nArgs!=2 && nArgs!=4)
		throw RuntimeException("Wrong number of arguments; found: "+boost::lexical_cast<std::string>(nArgs)+", expecting: 2 or 4", invocationExpression);
	intrusive_ptr<RingValue> R = runtimeEnv->evalArgAs<RingValue>(ARG(0));
  //	vector<string> v = runtimeEnv->evalArgAsListOf<StringValue>(ARG(1));
	intrusive_ptr<ListValue> l = runtimeEnv->evalArgAs<ListValue>(ARG(1));
	ListValue::ContainerType::size_type size = l->size();
  vector<symbol> v;
  for (unsigned long i=0; i<size; ++i)
  {
		if (const boost::intrusive_ptr<StringValue> s = boost::dynamic_pointer_cast<StringValue>(l->getValue(i)))
      v.push_back(symbol(s->theString));
    else if (const boost::intrusive_ptr<RecordValue> r = boost::dynamic_pointer_cast<RecordValue>(l->getValue(i)))
    {
      const boost::intrusive_ptr<StringValue> s = boost::dynamic_pointer_cast<StringValue>(r->getField("head"));
      const boost::intrusive_ptr<ListValue> inds = boost::dynamic_pointer_cast<ListValue>(r->getField("indices"));
      vector<long> indices;
      long tmp;
      ListValue::ContainerType::size_type NumIndices = inds->size();
      for (unsigned long j=0; j<NumIndices; ++j)
        if (IsConvertible(tmp, (boost::dynamic_pointer_cast<BigIntValue>(inds->getValue(j)))->theInteger))
          indices.push_back(tmp);
        else
          throw RuntimeException("All indices must fit into a machine-integer", ARG(1).exp);
      v.push_back(symbol(s->theString, indices));
    }
    else throw RuntimeException("List of symbols must have strings or records", ARG(1).exp);
  }
  if (nArgs==2)
    return Value::from(NewPolyRing(R->theRing, v));
	intrusive_ptr<MatrixValue> M = runtimeEnv->evalArgAs<MatrixValue>(ARG(2));
	intrusive_ptr<BigIntValue> GradingDim = runtimeEnv->evalArgAs<BigIntValue>(ARG(3));
  long d;
  if (!IsConvertible(d, GradingDim->theInteger))
		throw RuntimeException("invalid GradingDim", ARG(3).exp);
  const PPOrdering PPO = NewMatrixOrdering(v.size(), d, M->theMatrix);
	return new RingValue(NewPolyRing(R->theRing, NewPPMonoid(v, PPO)));
}
  //END_STD_BUILTIN_FUNCTION // no: variable number of args

//---- IDEAL -------------------------------------------------------

DECLARE_STD_BUILTIN_FUNCTION(elim, 2) { // AMB
	vector<RingElem> v1 = runtimeEnv->evalArgAsListOf<RingElemValue>(ARG(0));
	intrusive_ptr<IdealValue> I = runtimeEnv->evalArgAs<IdealValue>(ARG(1));
  ideal J = I->theIdeal;
  MakeUnique(J)->myElim(v1); // I->myElim(v1) ~~~> const
	return Value::from(J);
}
END_STD_BUILTIN_FUNCTION


DECLARE_STD_BUILTIN_FUNCTION(saturate, 2) { // AMB
	intrusive_ptr<IdealValue> I0 = runtimeEnv->evalArgAs<IdealValue>(ARG(0));
	intrusive_ptr<IdealValue> I1 = runtimeEnv->evalArgAs<IdealValue>(ARG(1));
  ideal J = I0->theIdeal;
  MakeUnique(J)->mySaturate(I1->theIdeal);
	return Value::from(J);
}
END_STD_BUILTIN_FUNCTION


DECLARE_STD_BUILTIN_FUNCTION(IdealOfPoints, 2) { // JAA 2013-01-19
	intrusive_ptr<RingValue> P = runtimeEnv->evalArgAs<RingValue>(ARG(0));
	intrusive_ptr<MatrixValue> pts = runtimeEnv->evalArgAs<MatrixValue>(ARG(1));
  return Value::from(IdealOfPoints(AsSparsePolyRing(P->theRing), pts->theMatrix));
}
END_STD_BUILTIN_FUNCTION


DECLARE_STD_BUILTIN_FUNCTION(TmpNBM, 3) { // AMB
	intrusive_ptr<RingValue> P = runtimeEnv->evalArgAs<RingValue>(ARG(0));
	intrusive_ptr<MatrixValue> Pts = runtimeEnv->evalArgAs<MatrixValue>(ARG(1));
	intrusive_ptr<MatrixValue> Tols = runtimeEnv->evalArgAs<MatrixValue>(ARG(2));
  vector<RingElem> QB;
  vector<RingElem> BB;
  vector<RingElem> AV;
  NBM_forC5(QB, BB, AV, AsSparsePolyRing(P->theRing), Pts->theMatrix, Tols->theMatrix);
  intrusive_ptr<RecordValue> rec(new RecordValue);
  rec->setField("QuotientBasis", Value::from(QB));
  rec->setField("AlmostVanishing", Value::from(AV));
  if (BB.empty())
    rec->setField("StableBBasisFound", Value::from(false));
  else
  {
    rec->setField("StableBBasisFound", Value::from(true));
    rec->setField("BBasis", Value::from(BB));
  }
  return rec;
}
END_STD_BUILTIN_FUNCTION

// DECLARE_STD_BUILTIN_FUNCTION(ClosePassingPoly, 3) { // AMB
// 	intrusive_ptr<RingValue> P = runtimeEnv->evalArgAs<RingValue>(ARG(0));
// 	intrusive_ptr<MatrixValue> Pts = runtimeEnv->evalArgAs<MatrixValue>(ARG(1));
// 	intrusive_ptr<MatrixValue> Tols = runtimeEnv->evalArgAs<MatrixValue>(ARG(2));
// 	intrusive_ptr<BigRatValue> MaxTol = runtimeEnv->evalArgAs<BigRatValue>(ARG(3));
//   ClosePassingPoly_forC5(AsSparsePolyRing(P->theRing), Pts->theMatrix, Tols->theMatrix);
//   return Value::from(ClosePassingPoly_forC5(AsSparsePolyRing(P->theRing), Pts->theMatrix, Tols->theMatrix);
// }
// END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(PreprocessPts, 2) { // JAA
	intrusive_ptr<MatrixValue> OrigPts = runtimeEnv->evalArgAs<MatrixValue>(ARG(0));
	intrusive_ptr<MatrixValue> epsilon = runtimeEnv->evalArgAs<MatrixValue>(ARG(1));
        ring R = BaseRing(OrigPts->theMatrix);
        vector< vector<RingElem> > NewPts;
        vector<long> weights;
        PreprocessPts_forC5("auto", NewPts, weights, OrigPts->theMatrix, epsilon->theMatrix);
//  NBM_forC5(QB, BB, AV, AsSparsePolyRing(P->theRing), Pts->theMatrix, Tols->theMatrix);
  intrusive_ptr<RecordValue> rec(new RecordValue);
  rec->setField("NewPoints", Value::from(NewDenseMat(R,NewPts)));
  rec->setField("weights", Value::from(weights));
  return rec;
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(PreprocessPtsGrid, 2) { // JAA
	intrusive_ptr<MatrixValue> OrigPts = runtimeEnv->evalArgAs<MatrixValue>(ARG(0));
	intrusive_ptr<MatrixValue> epsilon = runtimeEnv->evalArgAs<MatrixValue>(ARG(1));
        ring R = BaseRing(OrigPts->theMatrix);
        vector< vector<RingElem> > NewPts;
        vector<long> weights;
        PreprocessPts_forC5("grid", NewPts, weights, OrigPts->theMatrix, epsilon->theMatrix);
//  NBM_forC5(QB, BB, AV, AsSparsePolyRing(P->theRing), Pts->theMatrix, Tols->theMatrix);
  intrusive_ptr<RecordValue> rec(new RecordValue);
  rec->setField("NewPoints", Value::from(NewDenseMat(R,NewPts)));
  rec->setField("weights", Value::from(weights));
  return rec;
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(PreprocessPtsAggr, 2) { // JAA
	intrusive_ptr<MatrixValue> OrigPts = runtimeEnv->evalArgAs<MatrixValue>(ARG(0));
	intrusive_ptr<MatrixValue> epsilon = runtimeEnv->evalArgAs<MatrixValue>(ARG(1));
        ring R = BaseRing(OrigPts->theMatrix);
        vector< vector<RingElem> > NewPts;
        vector<long> weights;
        PreprocessPts_forC5("aggr", NewPts, weights, OrigPts->theMatrix, epsilon->theMatrix);
//  NBM_forC5(QB, BB, AV, AsSparsePolyRing(P->theRing), Pts->theMatrix, Tols->theMatrix);
  intrusive_ptr<RecordValue> rec(new RecordValue);
  rec->setField("NewPoints", Value::from(NewDenseMat(R,NewPts)));
  rec->setField("weights", Value::from(weights));
  return rec;
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(PreprocessPtsSubdiv, 2) { // JAA
	intrusive_ptr<MatrixValue> OrigPts = runtimeEnv->evalArgAs<MatrixValue>(ARG(0));
	intrusive_ptr<MatrixValue> epsilon = runtimeEnv->evalArgAs<MatrixValue>(ARG(1));
        ring R = BaseRing(OrigPts->theMatrix);
        vector< vector<RingElem> > NewPts;
        vector<long> weights;
        PreprocessPts_forC5("subdiv", NewPts, weights, OrigPts->theMatrix, epsilon->theMatrix);
//  NBM_forC5(QB, BB, AV, AsSparsePolyRing(P->theRing), Pts->theMatrix, Tols->theMatrix);
  intrusive_ptr<RecordValue> rec(new RecordValue);
  rec->setField("NewPoints", Value::from(NewDenseMat(R,NewPts)));
  rec->setField("weights", Value::from(weights));
  return rec;
}
END_STD_BUILTIN_FUNCTION


// DECLARE_STD_BUILTIN_FUNCTION(JanetBasis, 1) { // AMB + MA
// 	intrusive_ptr<IdealValue> I = runtimeEnv->evalArgAs<IdealValue>(ARG(0));
//   return Value::from(JanetBasis(I->theIdeal));
// }
// END_STD_BUILTIN_FUNCTION


//---- general ----

DECLARE_STD_BUILTIN_FUNCTION(TopLevelFunctions, 0) { // GL
	return runtimeEnv->topLevelFunctions();
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(CocoaPackagePath, 0) { // GL
	(void)runtimeEnv; // keeps the compiler happy (or, at least, silent ;-) )
	return Value::from(packageDir);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(VersionInfo, 0) { // AMB
	(void)runtimeEnv; // keeps the compiler happy (or, at least, silent ;-) )
  intrusive_ptr<RecordValue> rec(new RecordValue);
  rec->setField("CoCoALibVersion", Value::from(BuildInfo::version()));
  rec->setField("CoCoAVersion", Value::from(CoCoAVersion()));
  rec->setField("CompilationDate", Value::from(CompilationDate()));
  rec->setField("CompilationFlags", Value::from(BuildInfo::CompilationFlags()));
  rec->setField("CompilationDefines", Value::from(BuildInfo::CompilationDefines()));
  rec->setField("Compiler", Value::from(BuildInfo::compiler()));
  rec->setField("MachineIntNumBits", new BigIntValue(std::numeric_limits<unsigned int>::digits));
  rec->setField("MachineLongNumBits", new BigIntValue(std::numeric_limits<unsigned long>::digits));

  return rec;
}
END_STD_BUILTIN_FUNCTION

//---- manual CoCoAHelp ----

DECLARE_STD_BUILTIN_FUNCTION(ReloadMan, 0) { // AMB
	try {
    ostringstream os;
    OnlineHelp::ReloadMan(os);
    runtimeEnv->getOutputStream()->print(os.str())->flush();
	} catch (const std::exception& err) {
		throw RuntimeException(err.what(), invocationExpression->getBegin(), invocationExpression->getEnd());
	}
	return VoidValue::theInstance;
}
END_STD_BUILTIN_FUNCTION


  //----------------------------------------------------------------------
void RuntimeEnvironment::initBuiltInFunctions()
{
  BOOST_FOREACH(NameFunPair &p, builtIns())
		this->setTopLevelVar(p.first, p.second, VariableSlot::VSF_SystemProtected);
}
  //----------------------------------------------------------------------

} // namespace InterpreterNS
} // namespace CoCoA
